/* $Id: cgistream.c,v 1.13 2003/01/02 22:36:49 aa5779 Exp $ */
/*
	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "util.h"
#include "lists.h"
#include "interp.h"
#define CGISTREAM_IMP
#include "cgistream.h"
#undef putc

stream_class incgi_stream_class = {"incgi", ics_open, cgi_close, ics_read, 
									dummy_writestr, 
								    dummy_writetag, cgi_ioctl, dummy_getpos,
								    dummy_setpos, dummy_rewind
								 };

stream_class outcgi_stream_class = {"outcgi", ocs_open, cgi_close, dummy_read,
									ocs_writestr, dummy_writetag, cgi_ioctl,
									dummy_getpos, dummy_setpos, dummy_rewind,
									};
									
static int cgi_input_read;
static html_chunk form_chunk;

static char *strtok_x(char *data, char *delims)
{
	static char *next;
	char *ptr;
	char *base;

	if(data)
		next = data;
	if(!next)
		return NULL;
	base = next;
	ptr = next + strcspn(next, delims);
	next = *ptr ? ptr + 1 : NULL;
	*ptr = '\0';
	return base;
}

static void parse_url_encode(char *buffer)
{
	char *name = strtok_x(buffer, "=");
	char *value;
	dstring *val;
	while(name)
	{
		val = ds_create(NULL);
		value = strtok_x(NULL, "&;");
		if(!value)
			break;
		for(; *value; value++)
		{
			switch(*value)
			{
				case '+':
					ds_appendch(val, ' ');
					break;
				case '%':
				{
					int ch = toupper(*++value);
					int v = isalpha(ch) ? ch - 'A' + 10 : ch - '0';
					ch = toupper(*++value);
					v = (v << 4) | (isalpha(ch) ? ch - 'A' + 10 : ch - '0');
					ds_appendch(val, v);
					break;
				}
				default:
					ds_appendch(val, *value);
			}
		}
		set_html_param(&form_chunk, ds_create(name), val);
		name = strtok_x(NULL, "=");
	}
}

stream *ics_open(interpreter *i, char *name, char *fname, int writeable)
{
	char *req_method = getenv("REQUEST_METHOD");
	char *query;
	stream *ics_stream;

	ics_stream = ALLOC(stream);
	ics_stream->class = &incgi_stream_class;
	ics_stream->direction = stream_in;
	if(!cgi_input_read)
	{
		cgi_input_read = 1;
		if(!req_method || !strcmp(req_method, "GET") || !strcmp(req_method, "HEAD"))
		{
			query = getenv("QUERY_STRING");
			query = xstrdup(query ? query : "");
		}
		else if(!strcmp(req_method, "POST") || !strcmp(req_method, "PUT"))
		{
			int cnt = atoi(getenv("CONTENT_LENGTH"));
			if(strcmp(getenv("CONTENT_TYPE"), "application/x-www-form-urlencoded"))
				return ics_stream;
			query = xmalloc(cnt + 1);
			read(STDIN_FILENO, query, cnt);
			query[cnt] = '\0';
		}
		else query = xstrdup("");
		set_tag_name(&form_chunk, ds_create("QUERY"));
		parse_url_encode(query);
		free(query); 
	}
	return ics_stream;
}


DEFLIST(cgi_handler,
		char *mime_type;
		char *stream_type;
	);

static void destroy_cgi_handler(cgi_handler *h)
{
	free(h->mime_type);
}

void cgi_close(interpreter *i, stream *s)
{
	list_delete(s->data, (list_destroy_t)destroy_cgi_handler, sizeof(cgi_handler));
}

static int find_cgi_handler(cgi_handler *data, char *key)
{
	return !strcmp(data->mime_type, key);
}

static char *find_bymime(cgi_handler *list, char *mime)
{
	cgi_handler *h = list_find(list, mime, (list_predicate_t)find_cgi_handler);
	if(!h)
	{
		char *tmp = strchr(mime, '/');
		if(!tmp || !tmp[1])
			return NULL;
		tmp[1] = '*';
		tmp[2] = '\0';
		h = list_find(list, mime, (list_predicate_t)find_cgi_handler);
		if(!h)
			h = list_find(list, "*/*", (list_predicate_t)find_cgi_handler);
	}
	return h ? h->stream_type : NULL;
}

html_chunk *ics_read(interpreter *i, stream *s)
{
	if(HTML_CHUNK_TYPE(&form_chunk) == hct_none)
	{
		char *detect;
		char *mime = xstrdup(getenv("CONTENT_TYPE"));
		char *tmp = strchr(mime, ';');
		stream_class *cls;
		stream *ns;
		dstring *options = ds_create(NULL);

		if(tmp)
			*tmp = '\0';
		detect = find_bymime(s->data, mime);
		free(mime);
		cls = detect ? detect_stream_class(i, detect, &options) : &basic_stream_class;
		list_delete(s->data, (list_destroy_t)destroy_cgi_handler, sizeof(cgi_handler));
		ns = cls->open(i, "", DS_BODY(options), 1);
		s->class = ns->class;
		s->data = ns->data;
		s->direction = ns->direction;
		FREE(ns);
		return s->class->read_chunk(i, s);
	}
	return &form_chunk;
}

char *cgi_ioctl(interpreter *i, stream *s, char *ctl)
{
	char *sep;
	cgi_handler *nh;
	nh = ALLOC(cgi_handler);
	nh->mime_type = xstrdup(ctl);
	sep = strchr(nh->mime_type, '=');
	if(!sep || sep == ctl)
		recover_error("invalid MIME mapping: %s", ctl);
	nh->stream_type = sep + 1;
	s->data = list_add(s->data, nh);
	return "";
}

stream *ocs_open(interpreter *i, char *name, char *fname, int writeable)
{
	stream *s = ALLOC(stream);
	s->direction = writeable;
	s->class = &outcgi_stream_class;
	return s;
}
	
void ocs_writestr(interpreter *i, stream *s, char *str)
{
	if(*str)
	{
		if(!strncasecmp(str, "Content-Type:", 13))
		{
			char *mime = xstrdup(str + 13 + strspn(str + 13, " \t\n"));
			char *tmp = strchr(mime, ';');
			char *detect;
			cgi_handler *nc;

			if(tmp)
				*tmp = '\0';
			detect = find_bymime(s->data, mime);
			if(detect)
			{
				nc = ALLOC(cgi_handler);
				nc->stream_type = xstrdup(detect);
				nc->mime_type = "";
				list_delete(s->data, (list_destroy_t)destroy_cgi_handler, sizeof(cgi_handler));
				free(mime);
				s->data = nc;
			}
		}
		puts(str);
	}
	else
	{
		stream_class *cls;
		stream *ns;
		dstring *options = ds_create(NULL);

		putchar('\n');
		if(s->data)
			cls = detect_stream_class(i, ((cgi_handler *)s->data)->stream_type, &options);
		else
			cls = &html_stream_class;
		list_delete(s->data, (list_destroy_t)destroy_cgi_handler, sizeof(cgi_handler));
		ns = cls->open(i, "", DS_BODY(options), 1);
		s->class = ns->class;
		s->data = ns->data;
		s->direction = ns->direction;
	}
}



